Tolea Delia-Livia-Cristiana
334CB

#include "u8g.h"
#include "2048.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <avr/interrupt.h>
#include <avr/io.h>
#define F_CPU   16000000

#define ANIMATION_STEPS 2

uint8_t grid[SIZE][SIZE];
u8g_t u8g;
static char debounced;
static int8_t started;

static char values[12][5] = { "", "   2", "   4", "   8",
"  16", "  32", "  64", " 128",
" 256", " 512", "1024", "2048" };

static uint8_t move_stage;
static uint8_t move_type;
static uint8_t move_i, move_j;
static uint8_t speeds[SIZE][SIZE];


void initialize(void)
{
	memset(grid, 0, SIZE * SIZE * sizeof(uint8_t));
	new_block();
	new_block();
}

void new_block()
{
	uint8_t new_value, i, j;

	uint8_t r = rand() % 256;
	if (r > 230)
		new_value = 2;
	else
		new_value = 1;

	do
	{
		r = rand() % 16;
		i = r / SIZE;
		j = r % SIZE;
	} while (grid[i][j] != EMPTY);

	grid[i][j] = new_value;
}

void move_left(uint8_t i, uint8_t speed[SIZE])
{
	uint8_t c = 0;
	uint8_t j = 0;
	uint8_t u = 0;

	memset(speed, 0, SIZE);

	for (j = 0; j < SIZE; ++j)
	{
		// mutam casuta cu valoare din dreapta in stanga unde e un loc liber
		c = j;
		if (grid[i][j] == EMPTY)
		{
			while (grid[i][++c] == EMPTY && c < SIZE);
			if (c >= SIZE)     break;

			grid[i][j] = grid[i][c];
			grid[i][c] = EMPTY;

			speed[c] = c - j;
		}

		// unim 
		if (j > 0 && u == 0 && grid[i][j - 1] == grid[i][j])
		{
			u = 1;
			++grid[i][j - 1];
			grid[i][j] = EMPTY;
			--j;
			++speed[c];
		}
		else
			u = 0;
	}
}

// reflecta matricea astfel ca m[i][0] <-> m[i][3], m[i][1] <-> m[i][2]
void reflect(uint8_t m[SIZE][SIZE])
{
	uint8_t i, t;

	for (i = 0; i < SIZE; ++i)
	{
		t = m[i][0];
		m[i][0] = m[i][SIZE - 1];
		m[i][SIZE - 1] = t;
		t = m[i][1];
		m[i][1] = m[i][SIZE - 2];
		m[i][SIZE - 2] = t;
	}
}

// roteste matricea stanga/dreapta
void rotate(uint8_t m[SIZE][SIZE], uint8_t type)
{
	uint8_t i, j;
	uint8_t new_m[SIZE][SIZE];

	if (type == LEFT)
	{
		for (i = 0; i < SIZE; ++i)
			for (j = 0; j < SIZE; ++j)
				new_m[i][j] = m[j][SIZE - i - 1];
	}
	else
	{
		for (i = 0; i < SIZE; ++i)
			for (j = 0; j < SIZE; ++j)
				new_m[i][j] = m[SIZE - j - 1][i];
	}

	memcpy(m, new_m, SIZE * SIZE * sizeof(uint8_t));
}

void move(uint8_t type, uint8_t speeds[SIZE][SIZE])
{
	uint8_t i;

	switch (type)
	{
	case LEFT:
		break;
	case RIGHT:
		reflect(grid);
		break;
	case UP:
		rotate(grid, LEFT);
		break;
	case DOWN:
		rotate(grid, RIGHT);
		break;
	}

	memset(speeds, 0, SIZE * SIZE * sizeof(uint8_t));
	for (i = 0; i < SIZE; ++i)
		move_left(i, speeds[i]);

	switch (type)
	{
	case LEFT:
		break;
	case RIGHT:
		reflect(grid);
		reflect(speeds);
		break;
	case UP:
		rotate(grid, RIGHT);
		rotate(speeds, RIGHT);
		break;
	case DOWN:
		rotate(grid, LEFT);
		rotate(speeds, LEFT);
		break;
	}
}

uint8_t test(uint8_t type, uint8_t speeds[SIZE][SIZE])
{
	uint8_t temp[SIZE][SIZE];
	memcpy(temp, grid, SIZE * SIZE * sizeof(uint8_t));

	move(type, speeds);

	uint8_t res = memcmp(temp, grid, SIZE * SIZE * sizeof(uint8_t));
	memcpy(grid, temp, SIZE * SIZE * sizeof(uint8_t));

	return res;
}

// cheat function
void inc(void)
{
	uint8_t i, j;

	for (i = 0; i < SIZE; ++i)
		for (j = 0; j < SIZE; ++j)
			if (grid[i][j])
				++grid[i][j];
}

// evaluates the board
uint8_t evaluate(void)
{
	uint8_t i, j;
	uint8_t dummy[SIZE][SIZE];

	if (!(test(LEFT, dummy) ||
		test(RIGHT, dummy) ||
		test(UP, dummy) ||
		test(DOWN, dummy)))
		return LOST;

	for (i = 0; i < SIZE; ++i)
		for (j = 0; j < SIZE; ++j)
			if (grid[i][j] == 11)
				return WON;

	return NEITHER;
}

uint8_t get_value(uint8_t i, uint8_t j)
{
	return grid[i][j];
}

void start(void)
{
    started = 1;
    srand((unsigned)TCNT1);
    initialize();
}

void begin_move(uint8_t type)
{
    move_stage = 1;
    move_type = type;
    switch (type)
    {
        case LEFT:
            move_i = 0;
            move_j = -1;
            break;
        case RIGHT:
            move_i = 0;
            move_j = 1;
            break;
        case UP:
            move_i = -1;
            move_j = 0;
            break;
        case DOWN:
            move_i = 1;
            move_j = 0;
            break;
    }
}

ISR(PCINT2_vect)
{
    if ((PINC & (1 << PC0)) == 0)
    {
        if (debounced == 1)
        {
            debounced = 0;
            if (started < 1)
                start();
#ifdef CHEAT
            else
                inc();
#endif
        }
    }
    else
        debounced = 1;
}

ISR(PCINT1_vect)
{
    switch ((~PINB & ((1 << PB4) | (1 << PB5) | (1 << PB6) | (1 << PB7))))
    {
        case (1 << PB4):
        if (debounced == 1)
        {
            debounced = 0;
            if (started < 1)
                start();
            else if (move_stage == 0 && test(LEFT, speeds))
                begin_move(LEFT);
        }
        break;
        
        case (1 << PB5):
        if (debounced == 1)
        {
            debounced = 0;
            if (started < 1)
                start();
            else if (move_stage == 0 && test(DOWN, speeds))
                begin_move(DOWN);
        }
        break;
        
        case (1 << PB6):
        if (debounced == 1)
        {
            debounced = 0;
            if (started < 1)
                start();
            else if (move_stage == 0 && test(UP, speeds))
                begin_move(UP);
        }
        break;
        
        case (1 << PB7):
        if (debounced == 1)
        {
            debounced = 0;
            if (started < 1)
                start();
            else if (move_stage == 0 && test(RIGHT, speeds))
                begin_move(RIGHT);
        }
        break;
        
        default:
            debounced = 1;
    }
}

void u8g_setup(void)
{  
    u8g_Init8Bit(&u8g, &u8g_dev_st7920_128x64_4x_8bit,
                PN(0, 0), PN(0, 1), PN(0, 2), PN(0, 3), PN(0, 4), PN(0, 5), PN(0, 6), PN(0, 7), // DATA PORT              
                PN(1, 2),    // ENABLE
                U8G_PIN_NONE,
                U8G_PIN_NONE,
                PN(1, 0),    // RS / DI
                PN(1, 1),    // RW
                PN(1, 3));   // RESET
}

void io_init(void)
{
    DDRB &= ~((1 << PB4) | (1 << PB5) | (1 << PB6) | (1 << PB7));
    PORTB |= (1 << PB4) | (1 << PB5) | (1 << PB6) | (1 << PB7);
    
    DDRC &= ~(1 << PC0);
    PORTC |= (1 << PC0);
    
    PCICR |= (1 << PCIE0) | (1 << PCIE2);
    PCMSK1 |= (1 << PCINT12) | (1 << PCINT13) | (1 << PCINT14) | (1 << PCINT15);
    PCMSK2 |= (1 << PCINT16);
}

void timer1_init(void)
{
    TCCR1A &= ~((1 << COM1A0) | (1 << COM1A1) |     // OC0A dezactivat
                (1 << COM1B0) | (1 << COM1B1) |     // OC0B dezactivat
                (1 << WGM10) | (1 << WGM11));       
    TCCR1B &= ~((1 << WGM12) | (1 << WGM13) | (1 << CS11) | (1 << CS12));  // mod normal
    TCCR1B |=   (1 << CS10);                        // fara prescalar
}

void draw(void)
{    
    int i, j;

    switch (started)
    {
        case 0:
            u8g_SetDefaultForegroundColor(&u8g);
            u8g_SetFont(&u8g, u8g_font_helvB10);
            u8g_DrawStr(&u8g, 5, 25, "Press any button");
            u8g_DrawStr(&u8g, 40, 45, "to start");
            break;
            
        case -1:
            u8g_SetDefaultForegroundColor(&u8g);
            u8g_SetFont(&u8g, u8g_font_helvB10);
            u8g_DrawStr(&u8g, 25, 15, "YOU WON!");
            u8g_DrawStr(&u8g, 5, 35, "Press any button");
            u8g_DrawStr(&u8g, 35, 55, "to restart");
            break;
        
        case -2:
            u8g_SetDefaultForegroundColor(&u8g);
            u8g_SetFont(&u8g, u8g_font_helvB10);
            u8g_DrawStr(&u8g, 25, 15, "YOU LOST!");
            u8g_DrawStr(&u8g, 5, 35, "Press any button");
            u8g_DrawStr(&u8g, 35, 55, "to restart");        
            break;
        
        default:
            u8g_SetFont(&u8g, u8g_font_helvR08);
            
            for (i = 0; i < SIZE; ++i)
                for (j = 0; j < SIZE; ++j)
                    if (get_value(i, j) > 0)
                    {
                        u8g_SetDefaultForegroundColor(&u8g);                                
                        u8g_DrawBox(&u8g, 
                                    j * 32 + 1 + speeds[i][j] * move_stage * move_j * (32 / ANIMATION_STEPS),
                                    i * 16 + 1 + speeds[i][j] * move_stage * move_i * (16 / ANIMATION_STEPS),
                                    30, 14);
                        u8g_SetDefaultBackgroundColor(&u8g);
                        u8g_DrawStr(&u8g, 
                                    j * 32 + 4 + speeds[i][j] * move_stage * move_j * (32 / ANIMATION_STEPS),
                                    i * 16 + 12 + speeds[i][j] * move_stage * move_i * (16 / ANIMATION_STEPS),
                                    values[get_value(i, j)]);
                    }
            break;
    }
}

int main(void)
{
    sys_init();
    u8g_setup();
    io_init();
    timer1_init();
    started = 0;
    debounced = 1;
    
    sei();

    for ( ; ; )
    {   
        if (started == 1)
        {
            if (evaluate() == WON)
                started = -1;
            else if (evaluate() == LOST)
                started = -2;
        }
        
        u8g_FirstPage(&u8g);
        do
        {
            draw();
        } 
        while (u8g_NextPage(&u8g));
        
        if (move_stage > 0)
        {
            move_stage = (move_stage + 1) % ANIMATION_STEPS;
            if (move_stage == 0)
            {
                move(move_type, speeds);
                new_block();
            }
        }   
    }
	return 0;
}

